// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/driver/intime/os_bar.c $
// $Rev: 22965 $
// $Date: 2013-08-26 15:03:51 -0500 (Mon, 26 Aug 2013) $

// INtime 4 driver

#include "main.h"



//*****************************************************************************
static int _bar_reqion_io_acquire(os_bar_t* bar)
{
	int	ret;

	for (;;)	// A convenience loop.
	{
		if ((bar->flags & D0) == 0)
		{
			// The region is not I/O mapped.
			ret	= -EINVAL;
			break;
		}

		// INtime does not include a mechanism for associating a region
		// with a single device or a driver.
		bar->requested	= 1;
		ret				= 0;
		bar->vaddr		= (VADDR_T) bar->phys_adrs;
		break;
	}

	return(ret);
}




//*****************************************************************************
static int _bar_reqion_mem_acquire(os_bar_t* bar)
{
	int		ret;
	void*	vp;

	for (;;)	// A convenience loop.
	{
		if (bar->flags & D0)
		{
			// The region is not memory mapped.
			ret	= -EINVAL;
			break;
		}

		// INtime does not include a mechanism for associating a region
		// with a single device or a driver.
		bar->requested	= 1;
		vp				= MapRtPhysicalMemory(bar->phys_adrs, bar->size);

		if (vp == NULL)
		{
			// This BAR region request failed.
			ret	= EFAULT;
			break;
		}

		// All went well.
		ret			= 0;
		bar->vaddr	= (VADDR_T) vp;
		break;
	}

	return(ret);
}



//*****************************************************************************
unsigned long os_bar_pci_address(os_pci_t* pci, u8 index)
{
	u32	adrs;

	adrs	= pci->pcidev.IoSpace[index].start;
	return(adrs);
}



//*****************************************************************************
u32 os_bar_pci_flags(os_pci_t* pci, u8 index)
{
	u32	flags;

	flags	= pci->pcidev.dwBaseAddr[index];

	if (flags & 0x1)
		flags	&= 0x3;
	else
		flags	&= 0xF;

	return(flags);
}



//*****************************************************************************
u32 os_bar_pci_size(os_pci_t* pci, u8 index)
{
	u32	size;

	size	= pci->pcidev.IoSpace[index].size;
	return(size);
}



//*****************************************************************************
int os_bar_region_acquire(os_bar_t* bar)
{
	int	ret;

	if (bar->io_mapped)
		ret	= _bar_reqion_io_acquire(bar);
	else
		ret	= _bar_reqion_mem_acquire(bar);

	return(ret);
}



//*****************************************************************************
void os_bar_region_release(os_bar_t* bar)
{
	if (bar->requested == 0)
	{
	}
	else if (bar->io_mapped)
	{
		// INtime does not include a mechanism for associating a region
		// with a single device or a driver.
		bar->requested	= 0;
		bar->vaddr		= 0;
	}
	else
	{
		// INtime does not include a mechanism for associating a region
		// with a single device or a driver.

		if (bar->vaddr)
			FreeRtMemory((void*) bar->vaddr);	// Assume it works.

		bar->requested	= 0;
		bar->vaddr		= 0;
	}
}


