// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/driver/intime/os_exit.c $
// $Rev: 22965 $
// $Date: 2013-08-26 15:03:51 -0500 (Mon, 26 Aug 2013) $

// INtime 4 driver

#include "main.h"



//*****************************************************************************
static void _dev_threads_exit(void)
{
	int			i;
	RTHANDLE	mbox;
	os_msg_t	msg;
	char		name[64];
	int			qty		= ARRAY_ELEMENTS(gsc_global.dev_list) * GSC_DEVS_PER_BOARD;
	RTHANDLE	root;
	WORD		type	= 0;
	int			w;

	root	= os_root_get_root_process(NULL);

	if (root != BAD_RTHANDLE)
	{
		for (i = 0; i < qty; i++)
		{
			sprintf(name, "%s.%d", DEV_NAME, i);
			mbox	= LookupRtHandle(root, name, NO_WAIT);

			if (mbox == BAD_RTHANDLE)
				continue;

			memset(&msg, 0, sizeof(msg));
			msg.msg_type	= OS_MSG_TYPE_EXIT;
			msg.size		= OS_MSG_SIZE_EXIT;
			SendRtData(mbox, &msg, msg.size);

			for (w = 0; w < 1000; w++)	// Wait until the mailbox is deleted
			{
				type	= GetRtHandleType(mbox);

				if (type == MAILBOX_TYPE)
					os_time_tick_sleep(100);
				else
					break;
			}
		}
	}
}



//*****************************************************************************
static void _close_each_device(void)
{
	int	i;

	for (i = 0; i < ARRAY_ELEMENTS(gsc_global.dev_list); i++)
	{
		if (gsc_global.dev_list[i] == NULL)
			continue;

#if (GSC_DEVS_PER_BOARD > 1)

		{
			int	j;

			for (j = 0; j < GSC_DEVS_PER_BOARD; j++)
				gsc_close(&gsc_global.dev_list[i]->channel[j]);
		}

#else

		gsc_close(gsc_global.dev_list[i]);

#endif
	}
}



//*****************************************************************************
static void _service_threads_exit(void)
{
	int			i;
	RTHANDLE	mbox;
	os_msg_t	msg;
	WORD		type;
	int			w;

	for (i = 0; i < ARRAY_ELEMENTS(gsc_global.service_threads.data); i++)
	{
		if (gsc_global.service_threads.data[i].initialized == 0)
			continue;

		// Wait for the thread to finish working, if possible.

		for (w = 0; w < 1000; w++)
		{
			if (gsc_global.service_threads.data[i].initialized)
				os_time_tick_sleep(5);
			else
				break;
		}

		// Send the Exit message.

		memset(&msg, 0, sizeof(msg));
		msg.msg_type	= OS_MSG_TYPE_EXIT;
		msg.size		= OS_MSG_SIZE_EXIT;
		mbox			= gsc_global.service_threads.data[i].mbox;
		SendRtData(mbox, &msg, msg.size);

		// Allow time for the thread to finish working, if possible.

		for (w = 0; w < 1000; w++)
		{
			type	= GetRtHandleType(mbox);

			if (type == MAILBOX_TYPE)
				os_time_tick_sleep(5);
			else
				break;
		}
	}
}



//*****************************************************************************
void os_exit_driver(void)
{
	dev_data_t*	dev;
	int			i;

	for (;;)	// A convenience loop.
	{
		if ((gsc_global.driver_loaded == 0) && (gsc_global.dev_qty == 0))
		{
			// There is nothing for us to do.
			break;
		}

		os_sem_lock(&gsc_global.sem);

		_dev_threads_exit();
		_close_each_device();
		_service_threads_exit();

		for (i = 0; i < ARRAY_ELEMENTS(gsc_global.dev_list); i++)
		{
			dev	= gsc_global.dev_list[i];

			if (dev)
			{
				gsc_init_remove_device(dev);
				os_mem_data_free(dev);
			}
		}

		os_ioctl_reset();
		gsc_global.dev_qty	= 0;

		if (gsc_global.driver_loaded)
		{
			gsc_global.driver_loaded	= 0;
			printf(	"%s: driver version %s successfully removed.\n",
					DEV_NAME,
					GSC_DRIVER_VERSION);
		}

		os_sem_destroy(&gsc_global.service_threads.sem);
		break;
	}
}



//*****************************************************************************
void os_exit_initiate(int flags)
{
	int			i;
	int			limit;
	RTHANDLE	mbox;
	char		name[64];
	RTHANDLE	root;

	os_root_mbox_send_exit_msg(flags);

	if (flags & EXIT_UNCAT)
	{
		limit	= ARRAY_ELEMENTS(gsc_global.dev_list);
		limit	*= GSC_DEVS_PER_BOARD;
		limit	+= ARRAY_ELEMENTS(gsc_global.dev_list);
		root	= os_root_get_root_process(NULL);

		for (i = -1; i < limit; i++)
		{
			if (i >= 0)
				sprintf(name, "%s.%d", DEV_NAME, i);
			else
				strcpy(name, DEV_NAME);

			mbox	= LookupRtHandle(root, name, NO_WAIT);

			if (mbox == BAD_RTHANDLE)
				continue;

			UncatalogRtHandle(root, name);
		}
	}
}


