// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/driver/intime/os_init.c $
// $Rev: 23636 $
// $Date: 2013-10-22 11:25:51 -0500 (Tue, 22 Oct 2013) $

// INtime 4 driver

#include "main.h"



//*****************************************************************************
static void _service_msg_default(os_msg_t* msg)
{
	if (msg->mbox != BAD_RTHANDLE)
	{
		msg->status	= OS_MSG_STS_MSG_TYPE;
		SendRtData(msg->mbox, &msg, 1);
	}
}



//*****************************************************************************
static void _service_msg_close(os_service_data_t* ptr, WORD size, os_msg_t* msg)
{
	GSC_ALT_STRUCT_T*	alt	= ptr->alt;
	int					ret;

	if ((size == OS_MSG_SIZE_CLOSE) && (msg->size == OS_MSG_SIZE_CLOSE))
	{
		msg->status				= OS_MSG_STS_OK;
		ret						= os_close(alt);
		msg->u.close.results	= ret ? -1 : 0;
		msg->u.close.err_no		= -ret;
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
	{
		SendRtData(msg->mbox, msg, msg->size);
	}
}



//*****************************************************************************
static void _service_msg_ioctl_lg(os_service_data_t* ptr, WORD size, os_msg_t* msg)
{
	GSC_ALT_STRUCT_T*	alt	= ptr->alt;
	int					ret;
	void*				vp;

	if ((size == OS_MSG_SIZE_IOCTL_LG) && (msg->size == OS_MSG_SIZE_IOCTL_LG))
	{
		vp	= MapRtSharedMemory((RTHANDLE) msg->u.ioctl_lg.mem);

		if (vp)
		{
			vp	= (void*) ((u32) vp + msg->u.ioctl_lg.offset);
			ret	= os_ioctl(alt, msg->u.ioctl_lg.cmd, vp);
			msg->u.ioctl_sm.results	= ret ? -1 : 0;
			msg->u.ioctl_sm.err_no	= -ret;
			msg->status				= OS_MSG_STS_OK;
			FreeRtMemory(vp);
		}
		else
		{
			msg->status	= OS_MSG_STS_MEM_MAP;
			msg->size	= 1;
		}
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
static void _service_msg_ioctl_sm(os_service_data_t* ptr, WORD size, os_msg_t* msg)
{
	GSC_ALT_STRUCT_T*	alt		= ptr->alt;
	int					ret;
	s32					sz;

	if ((size >= OS_MSG_SIZE_IOCTL_SM_MIN)	&&
		(size <= OS_MSG_SIZE_IOCTL_SM_MAX)	&&
		(size == msg->size))
	{
		sz	= OS_IOCTL_SIZE_DECODE(msg->u.ioctl_sm.cmd);

		if (sz == (size - OS_MSG_SIZE_IOCTL_SM_MIN))
		{
			ret	= os_ioctl(alt, msg->u.ioctl_sm.cmd, msg->u.ioctl_sm.buf);
			msg->u.ioctl_sm.results	= ret ? -1 : 0;
			msg->u.ioctl_sm.err_no	= -ret;
			msg->status				= OS_MSG_STS_OK;
		}
		else
		{
			msg->status	= OS_MSG_STS_MSG_SIZE;
			msg->size	= 1;
		}
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
static void _service_msg_open(os_service_data_t* ptr, WORD size, os_msg_t* msg)
{
	GSC_ALT_STRUCT_T*	alt	= ptr->alt;
	int					ret;

	if ((size == OS_MSG_SIZE_OPEN) && (msg->size == OS_MSG_SIZE_OPEN))
	{
		msg->status			= OS_MSG_STS_OK;
		ret					= os_open(alt);
		msg->u.open.results	= ret ? -1 : 0;
		msg->u.open.err_no	= -ret;
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
static void _service_msg_read(os_service_data_t* ptr, WORD size, os_msg_t* msg)
{
	GSC_ALT_STRUCT_T*	alt	= ptr->alt;
	int					ret;
	void*				vp;

	if ((size == OS_MSG_SIZE_READ) && (msg->size == OS_MSG_SIZE_READ))
	{
		vp	= MapRtSharedMemory((RTHANDLE) msg->u.read.mem);

		if (vp)
		{
			vp	= (void*) ((u32) vp + msg->u.read.offset);
			msg->status			= OS_MSG_STS_OK;
			ret					= os_read(alt, vp, msg->u.read.count);
			msg->u.read.results	= (ret < 0) ? -1 : ret;
			msg->u.read.err_no	= (ret < 0) ? -ret : 0;
			FreeRtMemory(vp);
		}
		else
		{
			msg->status	= OS_MSG_STS_MEM_MAP;
			msg->size	= 1;
		}
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
static void _service_msg_write(os_service_data_t* ptr, WORD size, os_msg_t* msg)
{
	GSC_ALT_STRUCT_T*	alt	= ptr->alt;
	int					ret;
	void*				vp;

	if ((size == OS_MSG_SIZE_WRITE) && (msg->size == OS_MSG_SIZE_WRITE))
	{
		vp	= MapRtSharedMemory((RTHANDLE) msg->u.write.mem);

		if (vp)
		{
			vp	= (void*) ((u32) vp + msg->u.write.offset);
			msg->status				= OS_MSG_STS_OK;
			ret						= os_write(alt, vp, msg->u.write.count);
			msg->u.write.results	= (ret < 0) ? -1 : ret;
			msg->u.write.err_no		= (ret < 0) ? -ret : 0;
			FreeRtMemory(vp);
		}
		else
		{
			msg->status	= OS_MSG_STS_MEM_MAP;
			msg->size	= 1;
		}
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
static void _service_thread(void* arg)
{
	os_msg_t			msg;
	os_service_data_t*	ptr	= (void*) arg;
	WORD				size;

	for (;;)
	{
		ptr->busy	= 0;

		for (;;)	// Get a message.
		{
			size	= ReceiveRtData(ptr->mbox, &msg, 10000);

			if (size)
				break;
		}

		if ((size < 4) || (size != msg.size))
		{
			// The message is too small to process or its content is invalid.

			if (msg.mbox != BAD_RTHANDLE)
			{
				msg.status	= OS_MSG_STS_MSG_SIZE;
				SendRtData(msg.mbox, &msg, 1);
			}

			continue;
		}

		if ((size == OS_MSG_SIZE_EXIT) && (msg.msg_type == OS_MSG_TYPE_EXIT))
			break;

		switch (msg.msg_type)
		{
			default:

				_service_msg_default(&msg);
				continue;

			case OS_MSG_TYPE_CLOSE:

				_service_msg_close(ptr, size, &msg);
				continue;

			case OS_MSG_TYPE_IOCTL_LG:

				_service_msg_ioctl_lg(ptr, size, &msg);
				continue;

			case OS_MSG_TYPE_IOCTL_SM:

				_service_msg_ioctl_sm(ptr, size, &msg);
				continue;

			case OS_MSG_TYPE_OPEN:

				_service_msg_open(ptr, size, &msg);
				continue;

			case OS_MSG_TYPE_READ:

				_service_msg_read(ptr, size, &msg);
				continue;

			case OS_MSG_TYPE_WRITE:

				_service_msg_write(ptr, size, &msg);
				continue;
		}
	}

	DeleteRtMailbox(ptr->mbox);
	memset(ptr, 0, sizeof(ptr[0]));
}



//*****************************************************************************
static int _service_thread_start(int index)
{
	RTHANDLE			mbox;
	WORD				priority;
	os_service_data_t*	ptr		= &gsc_global.service_threads.data[index];
	int					ret		= 0;
	RTHANDLE			thread;

	mbox	= CreateRtMailbox(DATA_MAILBOX);

	if (mbox == BAD_RTHANDLE)
	{
		ret	= -ENOMEM;
	}
	else
	{
		thread		= GetRtThreadHandles(THIS_PROCESS);
		priority	= GetRtThreadPriority(thread);
		thread		= CreateRtThread(	priority + 1,
										_service_thread,
										4096,
										(void*) ptr);

		if (thread == BAD_RTHANDLE)
		{
			DeleteRtMailbox(mbox);
			ptr->mbox	= 0;
			ret			= -ENOMEM;
		}
		else
		{
			ptr->mbox			= mbox;
			ptr->initialized	= 1;
		}
	}

	return(ret);
}



//*****************************************************************************
static int _service_thread_aquire(void)
{
	int			i;
	int			index	= -1;
	int			lock;
	int			ret;

	lock	= os_sem_lock(&gsc_global.service_threads.sem);

	for (i = 0; i < ARRAY_ELEMENTS(gsc_global.service_threads.data); i++)
	{
		if (gsc_global.service_threads.data[i].busy)
			continue;

		if (gsc_global.service_threads.data[i].initialized == 0)
		{
			ret	= _service_thread_start(i);

			if (ret)
				continue;
		}

		gsc_global.service_threads.data[i].busy	= 1;
		index	= i;
		break;
	}

	if (lock == 0)
		os_sem_unlock(&gsc_global.service_threads.sem);

	return(index);
}



//*****************************************************************************
static void _dev_thread(void* arg)
{
	BOOL					cat			= FALSE;
	int						i;
	RTHANDLE				mbox		= BAD_RTHANDLE;
	char					name[64];
	os_msg_t				msg;
	os_dev_thread_data_t*	ptr			= (void*) arg;
	RTHANDLE				root;
	WORD					size;

	for (;;)
	{
		root	= os_root_get_root_process(NULL);

		if (root == BAD_RTHANDLE)
			break;

		mbox	= CreateRtMailbox(DATA_MAILBOX);

		if (mbox == BAD_RTHANDLE)
		{
			printf(	"%s: Failed to create Device Mailbox: %s\n",
					DEV_NAME,
					name);
			break;
		}

		sprintf(name, "%s.%d", DEV_NAME, ptr->index);
		cat	= CatalogRtHandle(root, mbox, name);

		if (cat == FALSE)
		{
			printf(	"%s: Failed to catalogc Device Mailbox: %s\n",
					DEV_NAME,
					name);
			break;
		}

		// Start processing messages.
		ptr->mbox	= mbox;

		for (;;)	// Process all messages.
		{
			for (;;)	// Get a message.
			{
				size	= ReceiveRtData(mbox, &msg, 10000);

				if (size)
					break;
			}

			if ((size == OS_MSG_SIZE_EXIT) && (msg.msg_type == OS_MSG_TYPE_EXIT))
				break;

			i	= _service_thread_aquire();

			if (i < 0)
			{
				msg.status	= OS_MSG_STS_THREAD;

				if (msg.mbox != BAD_RTHANDLE)
					SendRtData(msg.mbox, &msg, 1);

				continue;
			}

			gsc_global.service_threads.data[i].alt	= ptr->alt;
			SendRtData(gsc_global.service_threads.data[i].mbox, &msg, size);
		}

		break;
	}

	if (cat)
		UncatalogRtHandle(root, name);

	if (mbox != BAD_RTHANDLE)
		DeleteRtMailbox(mbox);

	memset(ptr, 0, sizeof(ptr[0]));
}



//*****************************************************************************
static int _dev_thread_start(GSC_ALT_STRUCT_T* alt, int index)
{
	WORD					priority;
	os_dev_thread_data_t*	ptr;
	int						ret;
	RTHANDLE				thread;

	ptr			= &gsc_global.dev_threads.data[index];
	ptr->alt	= alt;
	ptr->index	= index;
	thread		= GetRtThreadHandles(THIS_PROCESS);
	priority	= GetRtThreadPriority(thread);
	thread		= CreateRtThread(	priority + 1,
									_dev_thread,
									4096,
									(void*) ptr);

	if (thread == BAD_RTHANDLE)
		ret	= -ENOMEM;
	else
		ret	= 0;

	return(ret);
}



//*****************************************************************************
static int _dev_thread_acquire(GSC_ALT_STRUCT_T* alt, int index)
{
	int	i;
	int	lock;
	int	ret		= -ENOMEM;

	lock	= os_sem_lock(&gsc_global.dev_threads.sem);

	for (i = 0; i < ARRAY_ELEMENTS(gsc_global.dev_threads.data); i++)
	{
		if (gsc_global.dev_threads.data[i].in_use)
			continue;

		ret	= _dev_thread_start(alt, i);

		if (ret)
			continue;

		gsc_global.dev_threads.data[i].in_use	= 1;
		ret	= 0;
		break;
	}

	if (lock == 0)
		os_sem_unlock(&gsc_global.dev_threads.sem);

	return(ret);
}



#if (GSC_DEVS_PER_BOARD > 1)

//*****************************************************************************
static int _start_dev_threads(void)
{
	int	i;
	int	j;
	int	index	= 0;
	int	ret		= 0;

	os_sem_create(&gsc_global.dev_threads.sem);
	os_sem_create(&gsc_global.service_threads.sem);	// destroyed on driver exit.

	for (i = 0; i < ARRAY_ELEMENTS(gsc_global.dev_list); i++)
	{
		if (gsc_global.dev_list[i] == NULL)
			continue;

		for (j = 0; (ret == 0) && (j < GSC_DEVS_PER_BOARD); j++)
		{
			ret	= _dev_thread_acquire(&gsc_global.dev_list[i]->channel[j], index);
			index++;
		}
	}

	os_sem_destroy(&gsc_global.dev_threads.sem);
	return(ret);
}

#else

//*****************************************************************************
static int _start_dev_threads(void)
{
	int	i;
	int	index	= 0;
	int	ret		= 0;

	os_sem_create(&gsc_global.dev_threads.sem);
	os_sem_create(&gsc_global.service_threads.sem);	// destroyed on driver exit.

	for (i = 0; (ret == 0) && (i < ARRAY_ELEMEMNTS(gsc_global.dev_list)); i++)
	{
		if (gsc_global.dev_list[i] == NULL)
			continue;

		ret	= _dev_thread_acquire(gsc_global.dev_list[i], index);
		index++;
	}

	os_sem_destroy(&gsc_global.dev_threads.sem);
	return(ret);
}

#endif



//*****************************************************************************
int os_init_driver(void)
{
	dev_data_t*	dev		= NULL;
	PCIDEV		pcidev;
	int			ret		= 0;
	SYSINFO		si;
	int			size	= sizeof(dev_data_t);
	int			sts;

	sprintf(gsc_global.built, "%s, %s", __DATE__, __TIME__);
	printf(	"%s: driver loading: version %s, built %s\n",
			DEV_NAME,
			GSC_DRIVER_VERSION,
			gsc_global.built);

	PciInitialize(0);
	os_sem_create(&gsc_global.sem);

	// Compute time information.
	CopyRtSystemInfo(&si);
	gsc_global.tick_rate	= 1000 / si.NucleusTickInterval * si.KernelTickRatio;
	gsc_global.tick_us		= 1000000L / gsc_global.tick_rate;

	// Locate the devices and add them to our list.
	memset(&pcidev, 0, sizeof(pcidev));
	pcidev.wVendorId	= 0x10B5;	// PLX
	pcidev.wDeviceId	= 0;		// Find all PLX devices.
	pcidev.wDeviceIndex	= 0;		// Find the first device.

	for (;;)
	{
		if (dev == NULL)
		{
			dev	= os_mem_data_alloc(size);

			if (dev == NULL)
				break;
		}

		memset(dev, 0, size);
		dev->pci	= &dev->pci_data;
		memcpy(&dev->pci_data.pcidev, &pcidev, sizeof(pcidev));
		sts	= PciFindDevice(&dev->pci_data.pcidev);

		if (sts == FALSE)
			break;			// A device was not found.

		ret	= gsc_init_add_device(dev);

		if (ret > 0)
		{
			// The referenced device was added.
			printf("%s: device loaded: %s\n", DEV_NAME, dev->model);
			dev	= NULL;
		}
		else if (ret == 0)
		{
			// The referenced device was not added.
			// This is not an error condition.
		}
		else
		{
			// The referenced device was not added.
			// An error may have been reported.
		}

		pcidev.wDeviceIndex++;	// Find the next device.
	}

	if (dev)
	{
		os_mem_data_free(dev);
		dev	= NULL;
	}

	// Perform initialization following device discovery.

	for (;;)	// A convenience loop.
	{
		if (gsc_global.dev_qty <= 0)
		{
			ret	= -ENODEV;
			printf(	"%s: driver load failure: version %s, built %s\n",
					DEV_NAME,
					GSC_DRIVER_VERSION,
					gsc_global.built);
			break;
		}

		ret	= os_ioctl_init();

		if (ret)
			break;

		ret	= _start_dev_threads();

		if (ret)
			break;

		gsc_global.driver_loaded	= 1;
		printf(	"%s: driver loaded: version %s, built %s\n",
				DEV_NAME,
				GSC_DRIVER_VERSION,
				gsc_global.built);
		ret	= 0;
		break;
	}

	return(ret);
}


