// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/driver/intime/os_root.c $
// $Rev: 22965 $
// $Date: 2013-08-26 15:03:51 -0500 (Mon, 26 Aug 2013) $

// INtime 4 driver

#include "main.h"



// variables ******************************************************************

static	char		_root_buf[4096];	// MUST BE LARGE ENOUGH!!!
static	RTHANDLE	_root_mbox;			// Mailbox for DEV, not DEV.X
static	os_sem_t	_root_sem;



//*****************************************************************************
static int _fill_buffer(void)
{
	const char*	comma;
    int			i;
	int			j;

	for (;;)	// A convenience loop.
	{
		i   = sprintf(	_root_buf,
						"version: %s\n"
						"built: %s\n"
						"boards: %d\n"
						"models: ",
						GSC_DRIVER_VERSION,
						gsc_global.built,
						gsc_global.dev_qty);
		comma	= "";

		for (j = 0; j < ARRAY_ELEMENTS(gsc_global.dev_list); j++)
		{
			if (gsc_global.dev_list[j])
			{
				i		= strlen(_root_buf);
				i		= sprintf(	_root_buf + i,
									"%s%s",
									comma,
									gsc_global.dev_list[j]->model);
				comma	= ",";
			}
		}

		strcat(_root_buf, "\n");

#if defined(DEV_SUPPORTS_PROC_ID_STR)

		// This is intended to accommodate board identification
		// for those that have user jumpers.
		i		= strlen(_root_buf);
		i		= sprintf(_root_buf + i, "ids: ");
		comma	= "";

		for (j = 0; j < ARRAY_ELEMENTS(gsc_global.dev_list); j++)
		{
			if (gsc_global.dev_list[j])
			{
				i		= strlen(_root_buf);
				i		= sprintf(	_root_buf + i,
									"%s%s",
									comma,
									gsc_global.dev_list[j]->proc_id_str);
				comma	= ",";
			}
		}

		strcat(_root_buf, "\n");

#endif

		break;
	}

	i	= strlen(_root_buf);
    return(i);
}



//*****************************************************************************
static int _read_data(char* dst, int size)
{
	int	len;

	len	= _fill_buffer();

	if (size == 0)
	{
		len	= 0;
	}
	else if ((len  + 1) <= size)
	{
		strcpy(dst, _root_buf);
	}
	else
	{
		len	= size - 1;
		strncpy(dst, _root_buf, size);
		dst[len]	= 0;
	}

	return(len);
}



//*****************************************************************************
static void _service_msg_close(WORD size, os_msg_t* msg)
{
	if ((size == OS_MSG_SIZE_CLOSE) && (msg->size == OS_MSG_SIZE_CLOSE))
	{
		msg->status				= OS_MSG_STS_OK;
		msg->u.close.results	= 0;
		msg->u.close.err_no		= 0;
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
static void _service_msg_default(os_msg_t* msg)
{
	if (msg->mbox != BAD_RTHANDLE)
	{
		msg->status	= OS_MSG_STS_MSG_TYPE;
		SendRtData(msg->mbox, &msg, 1);
	}
}



//*****************************************************************************
static void _service_msg_open(WORD size, os_msg_t* msg)
{
	if ((size == OS_MSG_SIZE_OPEN) && (msg->size == OS_MSG_SIZE_OPEN))
	{
		msg->status			= OS_MSG_STS_OK;
		msg->u.open.results	= 0;
		msg->u.open.err_no	= 0;
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
static void _service_msg_read(WORD size, os_msg_t* msg)
{
	int		ret;
	void*	vp;

	if ((size == OS_MSG_SIZE_READ) && (msg->size == OS_MSG_SIZE_READ))
	{
		vp	= MapRtSharedMemory((RTHANDLE) msg->u.read.mem);

		if (vp)
		{
			vp	= (void*) ((u32) vp + msg->u.read.offset);
			ret					= _read_data(vp, msg->u.read.count);
			msg->status			= OS_MSG_STS_OK;
			msg->u.read.results	= (ret < 0) ? -1 : ret;
			msg->u.read.err_no	= (ret < 0) ? -ret : 0;
			FreeRtMemory(vp);
		}
		else
		{
			msg->status	= OS_MSG_STS_MEM_MAP;
			msg->size	= 1;
		}
	}
	else
	{
		msg->status	= OS_MSG_STS_MSG_SIZE;
		msg->size	= 1;
	}

	if (msg->mbox != BAD_RTHANDLE)
		SendRtData(msg->mbox, msg, msg->size);
}



//*****************************************************************************
RTHANDLE os_root_get_root_process(int* ret)
{
	RTHANDLE	root;

	root	= GetRtThreadHandles(ROOT_PROCESS);

	if (root == BAD_RTHANDLE)
	{
		printf(	"%s: Failed to get root process handle: %s\n",
				DEV_NAME,
				DEV_NAME);

		if (ret)
			ret[0]	= -EFAULT;
	}

	return(root);
}



//*****************************************************************************
int os_root_mbox_create(void)
{
	BOOL		cat;
	RTHANDLE	mbox;
	int			ret;
	RTHANDLE	root;

	for (;;)
	{
		root	= os_root_get_root_process(NULL);

		if (root == BAD_RTHANDLE)
		{
			ret	= -EFAULT;
			printf(	"%s: Can't access root process.\n", DEV_NAME);
			break;
		}

		mbox	= LookupRtHandle(root, DEV_NAME, NO_WAIT);

		if (mbox != BAD_RTHANDLE)
		{
			ret	= -EALREADY;
			printf(	"%s: The driver is already running: %s\n",
					DEV_NAME,
					DEV_NAME);
			break;
		}

		_root_mbox	= CreateRtMailbox(DATA_MAILBOX);

		if (_root_mbox == BAD_RTHANDLE)
		{
			ret	= -EFAULT;
			printf(	"%s: Failed to create Driver Mailbox: %s\n",
					DEV_NAME,
					DEV_NAME);
			break;
		}

		cat	= CatalogRtHandle(root, _root_mbox, DEV_NAME);

		if (cat == FALSE)
		{
			ret	= -EFAULT;
			printf(	"%s: Failed to catalogc Driver Mailbox: %s\n",
					DEV_NAME,
					DEV_NAME);
			DeleteRtMailbox(_root_mbox);
			_root_mbox	= BAD_RTHANDLE;
			break;
		}

		os_sem_create(&_root_sem);
		ret	= 0;
		break;
	}

	return(ret);
}



//*****************************************************************************
void os_root_mbox_send_exit_msg(int flags)
{
	RTHANDLE	mbox;
	os_msg_t	msg;
	int			ret;
	RTHANDLE	root;

	for (;;)	// A convenience loop.
	{
		root	= os_root_get_root_process(&ret);

		if (root == BAD_RTHANDLE)
			break;

		mbox	= LookupRtHandle(root, DEV_NAME, NO_WAIT);

		if (mbox == BAD_RTHANDLE)
			break;

		memset(&msg, 0, sizeof(msg));
		msg.msg_type	= OS_MSG_TYPE_EXIT;
		msg.size		= OS_MSG_SIZE_EXIT;
		SendRtData(mbox, &msg, msg.size);
		break;
	}
}



//*****************************************************************************
void os_root_mbox_service(void)
{
	int			lock	= -1;
	os_msg_t	msg;
	RTHANDLE	root;
	WORD		size;

	for (;;)
	{
		if (lock == 0)
			os_sem_unlock(&_root_sem);

		for (;;)
		{
			size	= ReceiveRtData(_root_mbox, &msg, 5000);

			if (size)
				break;
		}

		lock	= os_sem_lock(&_root_sem);

		if ((size < 4) || (size != msg.size))
		{
			// The message is too small to process or its content is invalid.

			if (msg.mbox != BAD_RTHANDLE)
			{
				msg.status	= OS_MSG_STS_MSG_SIZE;
				SendRtData(msg.mbox, &msg, 1);
			}

			continue;
		}

		if ((size == OS_MSG_SIZE_EXIT) && (msg.msg_type == OS_MSG_TYPE_EXIT))
			break;

		switch (msg.msg_type)
		{
			default:

				_service_msg_default(&msg);
				continue;

			case OS_MSG_TYPE_CLOSE:

				_service_msg_close(size, &msg);
				continue;

			case OS_MSG_TYPE_OPEN:

				_service_msg_open(size, &msg);
				continue;

			case OS_MSG_TYPE_READ:

				_service_msg_read(size, &msg);
				continue;
		}
	}

	root	= os_root_get_root_process(NULL);
	UncatalogRtHandle(root, DEV_NAME);
	DeleteRtMailbox(_root_mbox);
	os_sem_destroy(&_root_sem);
}


