// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/lib/intime/os_open.c $
// $Rev: 26099 $
// $Date: 2014-05-16 10:51:54 -0500 (Fri, 16 May 2014) $

// INtime 4 library

#include "main.h"



//*****************************************************************************
// returns err_no: 0 = ok, !=0 = error value
int os_open(int index, const char* base, os_file_t** fp)
{
	int			err_no	= 0;
	os_mbox_t*	mbox	= NULL;
	os_msg_t*	msg		= NULL;
	char		name[64];
	os_file_t*	ptr		= NULL;
	RTHANDLE	root;
	int			sem		= BAD_RTHANDLE;
	WORD		size;
	BOOL		sts;

	for (;;)
	{
		if ((index < -1) || (base == NULL) || (fp == NULL) ||
			(strlen(base) >= (sizeof(name) - 5)))
		{
			err_no	= EINVAL;
			break;
		}

		// Allocate memory for the  file structure.
		ptr	= malloc(sizeof(os_file_t));

		if (ptr == NULL)
		{
			err_no	= GetLastRtError();
			break;
		}

		memset(ptr, 0, sizeof(os_file_t));

		// Access the root process.
		root	= GetRtThreadHandles(ROOT_PROCESS);

		if (root == BAD_RTHANDLE)
		{
			err_no	= GetLastRtError();
			break;
		}

		// Access the designated mailbox.

		if (index == -1)
			strcpy(name, base);
		else
			sprintf(name, "%s.%d", base, index);

		ptr->dev_mbox	= LookupRtHandle(root, name, NO_WAIT);

		if (ptr->dev_mbox == BAD_RTHANDLE)
		{
			err_no	= GetLastRtError();
			break;
		}

		// Create the semaphore for the file structure.
		sem	= os_sem_create(&ptr->sem);

		if (sem)
		{
			errno	= sem;
			break;
		}

		// Acquire a mailbox for the receive message.
		err_no	= os_mbox_alloc(&mbox);

		if (err_no)
			break;

		// Allocate a message structure.
		err_no	= os_msg_alloc(&msg);

		if (err_no)
			break;

		// Initialize and send the message.
		memset(msg, 0, sizeof(os_msg_t));
		msg->status		= OS_MSG_STS_OK;
		msg->msg_type	= OS_MSG_TYPE_OPEN;
		msg->size		= OS_MSG_SIZE_OPEN;
		msg->mbox		= mbox->mbox;
		sts				= SendRtData(ptr->dev_mbox, msg, OS_MSG_SIZE_OPEN);

		if (sts == FALSE)
		{
			err_no	= GetLastRtError();
			break;
		}

		// Get and process the response.
		size	= ReceiveRtData(mbox->mbox, msg, WAIT_FOREVER);

		if ((size != OS_MSG_SIZE_OPEN) || (size != msg->size))
		{
			err_no	= EIO;
			break;
		}

		switch (msg->status)
		{
			default:

				err_no	= EBADMSG;
				break;

			case OS_MSG_STS_OK:

				ptr->val_ptr	= ptr;
				err_no			= msg->u.open.err_no;
				break;

			case OS_MSG_STS_MSG_SIZE:

				err_no	= EMSGSIZE;
				break;

			case OS_MSG_STS_MSG_TYPE:

				err_no	= EBADMSG;
				break;

			case OS_MSG_STS_MEM_MAP:

				err_no	= ENOMEM;
				break;

			case OS_MSG_STS_THREAD:

				err_no	= ENOEXEC;
				break;
		}

		break;
	}

	if (msg)
		os_msg_free(&msg);

	os_mbox_free(&mbox);

	if ((err_no) && (ptr))
	{
		if (sem == 0)
			os_sem_destroy(&ptr->sem);

		memset(ptr, 0, sizeof(ptr[0]));
		free(ptr);
		ptr	= NULL;
	}

	if (fp)
		fp[0]	= ptr;

	return(err_no);
}


