// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/SIO4%20and%20SIO8/SIO4_Linux_2.x.x_GSC_DN/utils/util_usc_hdlc_rx_adrs_ctrl.c $
// $Rev: 18173 $
// $Date: 2012-09-26 09:55:58 -0500 (Wed, 26 Sep 2012) $

#include "sio4_utils.h"



/******************************************************************************
*
*	Function:	sio4_usc_hdlc_rx_adrs_ctrl
*
*	Purpose:
*
*		Provide a visual wrapper for the SIO4_IOCTL_USC_HDLC_RX_ADRS_CTRL service.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the board to access. Ignore if < 0.
*
*		verbose	0: errors only, !0: all output
*
*		set		This is the value to apply.
*
*		get		The results are reported here. This may be NULL.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int sio4_usc_hdlc_rx_adrs_ctrl(int fd, int index, int verbose, s32 set, s32* get)
{
	char		buf[128];
	int			errs;
	const char*	ptr;

	if (verbose)
		gsc_label_index("USC HDLC Rx Address/Control", index);

	errs	= sio4_dsl_ioctl(fd, SIO4_IOCTL_USC_HDLC_RX_ADRS_CTRL, &set);

	switch (set)
	{
		default:

			errs++;
			ptr	= buf;
			sprintf(buf, "Unrecognized option: 0x%lX", (long) set);
			break;

		case -1:

			ptr	= "This service is not supported by this board.";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_OFF:

			ptr	= "No Address or Control fields";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_16:

			ptr	= "16-bits, Combined";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_24:

			ptr	= "24-bits, Combined";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_32:

			ptr	= "32-bits, Combined";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_EA_16:

			ptr	= "Extended Address, 16-bit Control";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_EA_24:

			ptr	= "Extended Address, 24-bit Control";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_EAC8:

			ptr	= "Extended Address, 8-bit w/ Extended Control";
			break;

		case SIO4_USC_HDLC_RX_ADRS_CTRL_EAC16:

			ptr	= "Extended Address, 16-bit w/ Extended Control";
			break;
	}

	if (verbose)
		printf("%s  (%s)\n", errs ? "FAIL <---" : "PASS", ptr);

	if (get)
		get[0]	= set;

	return(errs);
}


